import pandas as pd
from datetime import datetime, timedelta
from typing import Union

from src.synthetic_sample import sample_curve
from src.synthetic_sample.sample_curve import SampleCurve, CustomSampleCurve


def create_yearly_metadata_dict(period_definition: Union[tuple, list], annual_sales: int) -> dict:
    """ Creates dictionary containing each year and the total number of sales to attribute

    Args:
        period_definition: defines the time period to use
            if a tuple, expected to contain (start_date, end_date)
            if a list, expected to contain each requested year as an integer

    Returns:
        dictionary with metadata around the sales for each year. Keys are:
            start_date: first day to include for the year
            end_date: last day to include for the year
            sales: total number of sales for the year
    """
    if type(period_definition) == list:
        year_metadata_dict = {
            year: {}
            for year in period_definition
        }
    elif type(period_definition) == tuple:
        start_date = period_definition[0]
        end_date = period_definition[1]
        year_metadata_dict = {
            year: {
                "start_date": datetime(year, 1, 1),
                "end_date": datetime(year, 12, 31)
            }
            for year in range(start_date.year, end_date.year + 1)
        }
        year_metadata_dict[start_date.year]["start_date"] = pd.to_datetime(start_date)
        year_metadata_dict[end_date.year]["end_date"] = pd.to_datetime(end_date)
    else:
        raise NotImplementedError()

    for year in year_metadata_dict:
        year_dict = year_metadata_dict.get(year)
        if year_dict.get("start_date") is not None:
            start_date = year_dict["start_date"]
            end_date = year_dict["end_date"]
        else:
            start_date = datetime(year, 1, 1)
            end_date = datetime(year, 12, 31)
        days_of_year = (end_date - start_date)
        multiplier = days_of_year / (datetime(year, 12, 31) - datetime(year, 1, 1))
        year_dict["sales"] = multiplier * annual_sales

    return year_metadata_dict


def apply_curve_to_time_period(curve: CustomSampleCurve, period_definition: Union[tuple, list],
                               seasonal_distribution: dict, annual_sales: int, annual_growth_rate: float, curve_modifiers: list) -> pd.DataFrame:
    """ Given a curve and input parameters, apply the curve to the period to create the aggregated synthetic data

    Args:
        curve: CustomSampleCurve object to apply to the year
        period_definition: defines the tim eperiod to use
            if a tuple, expected to contain (start_date, end_date)
            if a list, expected to contain each requested year as an integer
        seasonal_distribution: dictionary indicating redistribution of output according to seasonal trends
        annual_sales: annual sales for the period
        annual_growth_rate: average annual growth rate (10% growth being 1.1)

    Returns:
        dataframe with the total number of sales for each period in the requested time period
    """
    yearly_metadata_dict = create_yearly_metadata_dict(period_definition, annual_sales)
    sales_df = curve.apply(yearly_metadata_dict, seasonal_distribution, curve_modifiers)

    if type(period_definition) == tuple:
        start_date = period_definition[0] - timedelta(days=(period_definition[0].weekday() + 1))
        end_date = period_definition[1] - timedelta(days=(period_definition[1].weekday() + 1))
        start_mask = sales_df[SampleCurve.period_column_name] >= start_date
        end_mask = sales_df[SampleCurve.period_column_name] <= end_date
        sales_df = sales_df[start_mask & end_mask]

    return sales_df.reset_index(drop=True)


def generate_synthetic_curves(period_type: sample_curve.PeriodType, period_definition: Union[tuple, list],
                              annual_growth_factor: float, curve_definition: dict, seasonal_distribution: dict,
                              annual_sales: int, packages_per_sale: float, quantity_per_sale: float, curve_modifiers: list) -> pd.DataFrame:
    """ Generates synthetic sales data for given json input

    Args:
        period_type: type of period to create aggregated data for
        period_definition: definition for the period, either a list of years or a tuple of (start_date, end_date)
        annual_growth_factor: average annual growth rate (10% growth being 1.1)
        curve_definition: dictionary to define the sales curve in terms of the cumulative percentage at the end of each period
        seasonal_distribution: dictionary indicating redistribution of output according to seasonal trends
        annual_sales: annual sales for the period
        packages_per_sale: average number of packages per sale
        quantity_per_sale: average quantity sold per sale

    Returns:
        dataframe with the total number of sales, packages, and quantity for each period in the requested time period
    """
    # Create synthetic sales curve
    custom_curve = sample_curve.CustomSampleCurve(period_type, curve_definition)
    sales_time_series_df = apply_curve_to_time_period(custom_curve, period_definition, seasonal_distribution,
                                                      annual_sales, annual_growth_factor, curve_modifiers)
    sales_time_series_df.rename(columns={SampleCurve.value_column_name: "sales"}, inplace=True)

    # Add derived values based on ratio
    sales_time_series_df["packages"] = (sales_time_series_df["sales"] * packages_per_sale).apply(int)
    sales_time_series_df["quantity"] = (sales_time_series_df["sales"] * quantity_per_sale).apply(int)

    return sales_time_series_df



# ====
# beverageBI.synthetic_sample_curve.py
# Class that defines a synthetic curve and methods to apply it over a time period
# ====


import holidays
import numpy as np
import pandas as pd
from datetime import datetime, timedelta
from enum import Enum
from typing import Union
from isoweek import Week


class AnchorType(Enum):
    """
    Enumeration of the available anchor types
    """
    holiday = "holiday"
    week_of_year = "week_of_year"
    month_of_year = "month_of_year"
    day_of_year = "day_of_year"


class PeriodType(Enum):
    """
    Enumeration of the available period types
    """
    week = "week"
    month = "month"


class SeasonDefinition(Enum):
    """
    Enumeration of the available seasons and the months they include
    """
    Q1 = [1, 2, 3]
    Q2 = [4, 5, 6]
    Q3 = [7, 8, 9]
    Q4 = [10, 11, 12]


class SampleCurveFeature:
    """Defines a single feature of an annual sales curve

    Attributes:
        anchor_type: What type of annual anchor is used to define the feature
            Possible values: "holiday", "week_of_year", "month_of_year", "day_of_year"
        anchor_point: Annual point to define the feature
            Possible values: (string) - holiday name, (int) - week or day of year
        anchor_value: Cumulative percent of total sales (0.0-1.0) completed by the end of the period of the anchor_point
        relative_start: Number of periods before anchor_point to define a relative cumulative percent value
        start_value: Cumulative percent of total sales (0.0-1.0) completed by the end of the period indicated by relative_start
        relative_end: Number of periods before anchor_point to define a relative cumulative percent value
        end_value: Cumulative percent of total sales (0.0-1.0) completed by the end of the period indicated by relative_end
    """

    def __init__(self, anchor_type: AnchorType, anchor_point: Union[str, int], anchor_value: Union[str, int],
                 period_type: PeriodType, relative_start: int = None, start_value: float = None,
                 relative_end: int = None,
                 end_value: float = None) -> None:
        """ Initializes the SampleCurveFeature
        """

        anchor_type = AnchorType(anchor_type)
        if anchor_type not in AnchorType:
            raise NotImplementedError

        self.anchor_type = anchor_type
        self.anchor_point = anchor_point
        self.anchor_value = anchor_value
        self.period_type = period_type
        self.relative_start = relative_start
        self.start_value = start_value
        self.relative_end = relative_end
        self.end_value = end_value

    def add_to_curve(self, year: int, year_df: pd.DataFrame) -> None:
        """ Add this feature to a dataframe for a specified year

        Args:
            year: Calendar year to add the feature for
            year_df: Dataframe to add the feature to, by modifying the percentages in value_column_name
        """
        self.__create_feature_for_year(year, year_df)
        if self.anchor_period is not None:
            self.__add_change_point(year_df, self.anchor_period, self.anchor_value)

            if self.start_value:
                self.__add_change_point(year_df, self.start_period, self.start_value)

            if self.end_value:
                self.__add_change_point(year_df, self.end_period, self.end_value)

    def __create_feature_for_year(self, year: int, year_df: pd.DataFrame) -> None:
        """  Identifies the dates to modify for the feature based on the specific calendar year

        Args:
            year: Calendar year to create the feature for
            year_df: Dataframe of rows for each period, to determine relative dates
        """
        if self.anchor_type == AnchorType.holiday:
            # Get Holiday Period
            holiday_date = holidays.UnitedStates(years=year).get_named(self.anchor_point)[0]
            holiday_date = pd.to_datetime(holiday_date)
            if self.period_type == PeriodType.week:
                self.anchor_period = holiday_date - timedelta(days=(holiday_date.weekday() + 1))
            if self.period_type == PeriodType.month:
                self.anchor_period = pd.Timestamp(holiday_date.year, holiday_date.month, 1)

        elif self.anchor_type == AnchorType.week_of_year:
            if self.period_type == PeriodType.month:
                raise ValueError("Cannot apply week_of_year feature to monthly curve")
            try:
                # If week has 53 years, adjust for extra week (First week of Q4)
                if Week(year, 53).week == 53:
                    if self.anchor_point >= 42:
                        self.anchor_point += 1

                date = year_df.loc[self.anchor_point - 1][SampleCurve.period_column_name]
                self.anchor_period = date
            except KeyError:
                self.anchor_period = None

        elif self.anchor_type == AnchorType.month_of_year:
            if self.period_type == PeriodType.week:
                raise ValueError("Cannot apply month_of_year feature to weekly curve")
            try:
                date = year_df.loc[self.anchor_point - 1][SampleCurve.period_column_name]
                self.anchor_period = date
            except KeyError:
                self.anchor_period = None

        elif self.anchor_type == AnchorType.day_of_year:
            date_string = f"{year}-{self.anchor_point}"
            date = pd.to_datetime(date_string, format="%Y-%j")
            # TODO add monthly handling capabilities
            if self.period_type == PeriodType.week:
                self.anchor_period = date - timedelta(days=(date.weekday() + 1))
            else:
                raise NotImplementedError("day_of_year features not implemented for monthly curves")

        if self.relative_start:
            if (self.relative_start >= 0) | (type(self.relative_start) != int):
                raise ValueError("relative_start must be a negative integer")
            if self.period_type == PeriodType.week:
                self.start_period = self.anchor_period + timedelta(days=self.relative_start * 7)
            elif self.period_type == PeriodType.month:
                self.start_month = self.anchor_period + timedelta(months=self.relative_start)

        if self.relative_end:
            if (self.relative_end <= 0) | (type(self.relative_end) != int):
                raise ValueError("relative_end must be a positive integer")
            if self.period_type == PeriodType.week:
                self.end_period = self.anchor_period + timedelta(days=self.relative_end * 7)
            elif self.period_type == PeriodType.month:
                self.start_month = self.anchor_period + timedelta(months=self.relative_end)

    @staticmethod
    def __add_change_point(year_df: pd.DataFrame, period: datetime, value: float) -> None:
        """ Updates the percentage of the value_column_name column in year_df for the record where the period_column_name column equals period

        Args:
            year_df: Dataframe to modify, by changing the percentage in value_column_name
            period: value in the period_column_name column for the record to modify
            value: percentage to update the value_column_name to
        """
        if value:
            period_mask = year_df[SampleCurve.period_column_name] == period
            year_df.loc[period_mask, SampleCurve.value_column_name] = value


class SampleCurve:
    """ Defines an annual sales curve to generate synthetic data

    Args:
        period_type: name of the column to designate the period
        features_list: list of SampleCurveFeature objects to make up the curve
    """
    value_column_name = "value"
    period_column_name = "period"

    def __init__(self, period_type: PeriodType, features_list=None):
        """ Initializes a SampleCurve for the period_type with the specific features
        """
        self.period_type = period_type
        if features_list is not None:
            self.curve_features = features_list

    def create_year(self, year: int) -> pd.DataFrame:
        """ Creates a dataframe for each period_type period for the year with null values for each row

        Args:
            year: Calendar year to create a dataframe for

        Returns:
            Dataframe with each period in the year as a row
        """
        start_date = pd.Timestamp(year, 1, 1)
        end_date = pd.Timestamp(year + 1, 1, 1) + timedelta(days=-1)

        # Create dataframe with a row for each period in the year
        year_df = pd.DataFrame()
        if self.period_type == PeriodType.week:
            # First week has the first Thursday of the year
            # If first day of year comes before Thursday, roll the start date back to include that Monday in the date range
            if start_date.dayofweek in [x for x in range(2, 4)]:
                start_date -= timedelta(days=7)

            # Last week is the week before the first Thursday of the next year
            if end_date.dayofweek < 3:
                end_date += timedelta(days=-7)

            year_df[self.period_column_name] = pd.date_range(start=start_date, end=end_date, freq="W-MON")
            year_df = year_df.reset_index()
            year_df = year_df.rename(columns={"index": "week_number"})
            year_df["week_number"] += 1

        elif self.period_type == PeriodType.month:
            year_df[self.period_column_name] = pd.date_range(start=start_date, end=end_date, freq="MS")

        # Add the value column with nulls for each row
        year_df[self.value_column_name] = np.NaN

        return year_df

    def apply(self, yearly_metadata_dict: dict, seasonal_distribution: dict, modifiers_list: list) -> pd.DataFrame:
        """ Applies the curve to each year specified in yearly_metadata_dict with the associated parameters

        Args:
            yearly_metadata_dict: dictionary with each year to compute as a key and a dictionary with the following as values
                - start_date: date to include in the first period
                - end_date: date to include in the last period
                - sales: total number of sales to distribute
            seasonal_distribution: dictionary with the percentage of sales in each quarter

        Returns:
            dataframe with sales by period for the entire span covered by yearly_metadata_dict
        """
        # Create each yearly curve separately and append as a dataframe to annual_curves
        annual_curves = []
        for year in yearly_metadata_dict:
            # Create the empty dataframe for the year
            year_dict = yearly_metadata_dict.get(year)
            year_df = self.create_year(year)

            # Add each feature to the curve
            for feature in self.curve_features:
                feature.add_to_curve(year, year_df)

            # Fill in intermediary values with cumulative sum
            year_df = year_df.interpolate()

            # Reduce cumulative sum to individual period values
            year_df[self.value_column_name] = year_df[self.value_column_name].diff().fillna(
                year_df[self.value_column_name])

            # Redistribute percentages according to seasonal_distribution
            for quarter in seasonal_distribution:
                mask = year_df[self.period_column_name].dt.month.isin(SeasonDefinition[quarter].value)
                unscaled_percent = year_df.loc[mask, self.value_column_name].sum()
                scaling_factor = seasonal_distribution[quarter] / unscaled_percent
                year_df.loc[mask, self.value_column_name] = year_df.loc[mask, self.value_column_name] * scaling_factor

            # Filter out all periods that were not included
            for date_cutoff in ["start_date", "end_date"]:
                if year_dict.get(date_cutoff) is not None:
                    if self.period_type == PeriodType.week:
                        cutoff_period = year_dict[date_cutoff] - timedelta(days=(year_dict[date_cutoff].weekday() + 1))
                    elif self.period_type == PeriodType.month:
                        cutoff_period = pd.Timestamp(year_dict[date_cutoff].year, year_dict[date_cutoff].month, 1)
                    else:
                        raise NotImplementedError(f"Only week and month PeriodTypes are implemented")
                    if date_cutoff == "start_date":
                        cutoff_mask = year_df[self.period_column_name] >= cutoff_period
                    else:
                        cutoff_mask = year_df[self.period_column_name] <= cutoff_period
                    year_df = year_df[cutoff_mask]

            # If covid modifier is requested, apply that scaling
            if "covid" in modifiers_list:
                year_df = self.apply_covid_distribution(year_df)

            # Distribute total sales amount
            normalize_sum = year_df[self.value_column_name].sum()
            year_df["value"] = (year_df["value"] / normalize_sum) * year_dict["sales"]

            # Apply smoothing
            year_df[self.value_column_name] = year_df[self.value_column_name].ewm(span=2).mean()
            year_df[self.value_column_name] = year_df[self.value_column_name].astype(int)

            annual_curves.append(year_df)
            
        # Combine all of the dataframes and return
        annual_sales_curve = pd.concat(annual_curves)

        return annual_sales_curve

    def apply_covid_distribution(self, year_df):
        start_date = datetime(2020,3,26)
        end_date = datetime(2021,9,1)
        covid_scaling_factor = 1.333
        mask_covid = (year_df[self.period_column_name] >= start_date) & (year_df[self.period_column_name] < end_date)
        year_df.loc[mask_covid, self.value_column_name] = year_df.loc[mask_covid, self.value_column_name] * covid_scaling_factor

        return year_df


class CustomSampleCurve(SampleCurve):
    """ Enables creation of creating a custom sales curve by passing in a list of metadata describing SalesCurveFeatures
    """

    def __init__(self, period_type, def_dict_list):
        """ Initializes a CustomSampleCurve using a list of dictionaries

        Args:
            period_type: type of periods to make up the curve
            def_dict_list: list of dictionaries defining each feature, where each must have keys corresponding
                           to the SampleCurveFeature attributes:
                Required: anchor_type, anchor_point, anchor_value, period_type
                Optional: relative_start, start_value, relative_end, end_value
        """
        # Compile the full list of SampleCurveFeatures from the provided dictionary
        features = []
        for def_dict in def_dict_list:
            features.append(
                SampleCurveFeature(anchor_type=def_dict.get("anchor_type"),
                                   anchor_point=def_dict.get("anchor_point"),
                                   anchor_value=def_dict.get("anchor_value"),
                                   period_type=def_dict.get("period_type"),
                                   relative_start=def_dict.get("relative_start"),
                                   start_value=def_dict.get("start_value"),
                                   relative_end=def_dict.get("relative_end"),
                                   end_value=def_dict.get("end_value"))
            )
        # Initialize the curve with full list of features
        SampleCurve.__init__(self, period_type, features)



import numpy as np
import pandas as pd
import random
from datetime import datetime
from dateutil.relativedelta import relativedelta
from scipy.stats import skewnorm

WEIGHT_LABEL = "weight"
MAX_PERCENTAGE_TO_UPDATE = 0.02


def get_annualized_integer(start_date: datetime, end_date: datetime, total_value: int):
    """ Annualized the total_value for the entire period in defined by start_date, end_date

    Args:
        start_date: first date of the period
        end_date: last date of the period
        total_value: integer total for the entire period

    Returns:
        annualized integer value
    """
    number_of_years = relativedelta(end_date, start_date).years
    annualized_value = int(total_value / number_of_years)

    return annualized_value


def get_sales_ratios(default_packages_per_sale: float, default_quantity_per_package: float, annual_sales: int = None,
                     annual_packages: int = None, annual_quantity: int = None) -> tuple:
    """ Given any combination of annual_sales, annual_packages, and annual_quantity, generates the ratios of packages:sales and quantity:packages

    Args:
        default_packages_per_sale: packages to sales ratio to use as a default
        default_quantity_per_package: quantity to packages ratio to use as a default
        annual_sales: *optional* number of sales for the entire period
        annual_packages: *optional* number of packages for the entire period
        annual_quantity: *optional* quantity for the entire period

    Returns:
        tuple of (annual_sales, packages_per_sale, quantity_per_package)

    Raises:
        ValueError: One of annual_sales, annual_packages, or annual_quantity must be provided
    """
    # We calculate off of annual_sales if possible
    if annual_sales is not None:
        # If annual_packages is provided use it to determine the packages_per_sale
        if annual_packages is not None:
            packages_per_sale = annual_packages / annual_sales
            # If annual_quantity is provided, use it to determine quantity_per_package, otherwise use a default
            if annual_quantity is not None:
                quantity_per_sale = annual_quantity / annual_sales
            else:
                quantity_per_sale = default_quantity_per_package * packages_per_sale
        # If annual_packages is not provided, use a default
        else:
            packages_per_sale = default_packages_per_sale
            # If total quantity is provided, calculate quantity_per_package using the default packages_per_sale
            if annual_quantity is not None:
                quantity_per_sale = annual_quantity / annual_sales
            # Otherwise use a default
            else:
                quantity_per_sale = default_quantity_per_package * packages_per_sale
    # If annual_sales is not provided, base the other metrics off of a default packages_per_sale value
    else:
        packages_per_sale = default_packages_per_sale
        # If annual_packages is given, use it and a default packages_per_sale to determine annual_sales
        if annual_packages is not None:
            annual_sales = int(annual_packages / packages_per_sale)
            # If annual_quantity is given, use it and annual_packages to determine quantity_per_package
            if annual_quantity is not None:
                quantity_per_sale = annual_quantity / annual_sales
            # Otherwise use a default
            else:
                quantity_per_sale = default_quantity_per_package * packages_per_sale
        else:
            # If annual_quantity is given, use default ratios to get packages_per_sale and annual_sales
            if annual_quantity is not None:
                quantity_per_sale = default_quantity_per_package * default_packages_per_sale
                packages_per_sale = default_packages_per_sale
                annual_sales = int((annual_quantity / default_quantity_per_package) / default_packages_per_sale)
            else:
                raise ValueError("One of annual_sales, annual_packages, or annual_quantity must be provided")

    return annual_sales, packages_per_sale, quantity_per_sale


def create_column_from_dist(df: pd.DataFrame, dist_dict: dict, column_name: str) -> None:
    """ Given a dataframe, create a new column `column_name` using the provided distribution

    Args:
        df: dataframe to add the column to
        dist_dict: dictionary of items and their relative weighting
        column_name: name of the updated column
    """
    # Normalize values of relative weights
    dist_dict = normalize_dist_dict(dist_dict)

    # Create columns for random numbers and new values
    df["temp_value"] = np.random.random(df.shape[0])
    df[column_name] = None
    # Check for random values that fall within a probability band defined by the weight
    # For each item check for records that fall between the previous and new probability range limits
    p = 0
    for item in dist_dict:
        mask_lower = df["temp_value"] >= p
        mask_upper = df["temp_value"] <= (p + dist_dict[item])
        # Add the specified item to those records
        df.loc[mask_lower & mask_upper, column_name] = item
        # Add this items probability to the range for the next item
        p += dist_dict[item]
    # Drop temporary column
    df.drop(columns=["temp_value"], inplace=True)


def normalize_dist_dict(raw_dist_dict: dict) -> dict:
    """ Given a dictionary of items and relative weights, modify the weights to normalize to 1

    Args:
        raw_dist_dict:  dictionary of items as keys and their weighting as the value

    Returns: updated dictionary normalized to 1 or an empty dictionary if None is passed
    """
    normalized_dict = {}
    if raw_dist_dict:
        # Get value to normalize
        normalized_value = sum(raw_dist_dict.values())
        # Divide into each value for each key
        normalized_dict = {
            key: raw_dist_dict[key] / normalized_value
            for key in raw_dist_dict
        }

    return normalized_dict


def create_daily_dist_dict(month_timestamp: pd.Timestamp, weekly_distribution: dict, weekday_distribution: dict) -> dict:
    """ For a given month, provides the percentage distribution across each day of the month

    Args:
        month_timestamp: timestamp in the month of interest
        weekly_distribution: dictionary indicating the probability of sales by week of the month
            -1 indicates last week of the month
        weekday_distribution: dictionary indicating the probability of sales by day of the week

    Returns:
        dictionary with probabilities for every day of the month
    """
    # Get full list of weeks for month
    start_date = month_timestamp.replace(day=1)
    end_date = start_date + relativedelta(months=1) + relativedelta(days=-1)
    date_range = pd.date_range(start=start_date, end=end_date, freq="W-MON")
    date_range = [date for date in date_range if date.month == month_timestamp.month]

    # Apply distribution, create dict of weeks and their relative weights
    weekly_dist_dict = {}
    # Set probability for each week based on weekly_distribution
    for index in range(0, len(date_range)):
        weekly_dist_dict[date_range[index]] = weekly_distribution.get(index)
        # If -1 is included in the keys, apply it on the last iteration of the loop
        if (-1 in weekly_distribution.keys()) & (index == len(date_range) - 1):
            weekly_dist_dict[date_range[index]] = weekly_distribution[-1]
    # Create a dataframe from the dictionary
    weekly_dist_df = pd.DataFrame.from_dict(weekly_dist_dict, orient='index', columns=[WEIGHT_LABEL])
    # If no values are included, set the distribution to be flat
    if weekly_dist_df[WEIGHT_LABEL].isnull().all():
        weekly_dist_df[WEIGHT_LABEL] = 1
    # Otherwise, set the distribution across non-specified weeks to be flat
    else:
        # Remove probability of all provided weights
        fill_value = 1 - weekly_dist_df[WEIGHT_LABEL].sum()
        # Fill value evenly across other values
        fill_value = fill_value / (len(weekly_dist_df) - 1)
        mask = weekly_dist_df[WEIGHT_LABEL].isnull()
        weekly_dist_df.loc[mask, WEIGHT_LABEL] = fill_value

    # For each week, create a distribution across the individual days
    daily_dist_dict = {}
    for index, row in weekly_dist_df.iterrows():
        # Create a date range for the individual week
        week_end = index + relativedelta(days=6)
        week_dates = pd.date_range(index, week_end)
        # Create a distribution dictionary by multiplying the week's weight with the weekdays weight for each day
        weekday_dist = {
            week_dates[weekday_index].date(): weekday_distribution[weekday_index] * row[WEIGHT_LABEL]
            for weekday_index in weekday_distribution
        }
        daily_dist_dict.update(weekday_dist)

    # Normalize values for the entire month
    normalized_daily_dist_dict = normalize_dist_dict(daily_dist_dict)

    return normalized_daily_dist_dict


def add_int_column_from_distribution(dataframe: pd.DataFrame, column_name: str, mean_value: float, skewness: float,
                                     exponential: float, floor_column: str = None, floor_value: int = None) -> None:
    """ Given a dataframe add a new int column from the distribution defined by mean_value, skewness, and exponential

    Args:
        dataframe: dataframe to add the column to
        column_name: name of column to add
        mean_value: average value for the new column
        skewness: skewness of the distribution for the new column
        exponential: exponential for the new column before fitting to mean
        floor_column: column to indicate the minimum value that the new column can contain
        floor_value: minimum value that the new column can contain
    """
    # Uses https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.skewnorm.html to pull from a random distribution
    dataframe[column_name] = skewnorm.rvs(a=skewness, size=len(dataframe))

    # Shifts outputs so the minimum value is 1
    dataframe[column_name] = dataframe[column_name] - dataframe[column_name].min() + 1
    # Apply exponential to curve
    dataframe[column_name] = dataframe[column_name] ** exponential
    # Scale curve to have a mean as provided
    dataframe[column_name] = (dataframe[column_name] * (mean_value / dataframe[column_name].mean()))
    # Convert to discrete integers
    dataframe[column_name] = dataframe[column_name].astype(int)
    # Handle floor cases
    if floor_column:
        mask = dataframe[column_name] < dataframe[floor_column]
        dataframe.loc[mask, column_name] = dataframe.loc[mask, floor_column]
    if floor_value:
        mask = dataframe[column_name] < floor_value
        dataframe.loc[mask, column_name] = floor_value

    # Add 1 to random records to add noise until desired mean is acheived
    dataframe = dataframe.reset_index(drop=True)
    while dataframe[column_name].mean() < mean_value:
        # Determine number of records to update based on the current delta from desired mean
        delta = mean_value - dataframe[column_name].mean()
        number_to_change = int(len(dataframe) * delta/2) + 1
        # Check against a cap on number of records to update at once (to add noise rather than a general shift)
        if number_to_change > len(dataframe) * MAX_PERCENTAGE_TO_UPDATE:
            number_to_change = int(len(dataframe) * MAX_PERCENTAGE_TO_UPDATE)
        # Identify number_to_change records to add 1 to and update them all at once
        records_to_change = set([random.randint(0, len(dataframe)) for x in range(0, number_to_change)])
        mask = dataframe.index.isin(records_to_change)
        dataframe.loc[mask, column_name] = dataframe.loc[mask, column_name] + 1
